/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: VLTIfileUtil.c,v 1.6 2004/10/07 13:29:09 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
#define _POSIX_SOURCE_1
#include "vltPort.h"

/* $Id: VLTIfileUtil.c,v 1.6 2004/10/07 13:29:09 jdejong Exp $ */

/*----------------------------------------------------------------------- */
/* FITS file utility routines                                             */
/*----------------------------------------------------------------------- */
#include "VLTIfileUtil.h"

int append_tables(char *infile, char *outfile)
/*----------------------------------------------------------------------- */
/*  Append all tables from one file onto another.                         */
/*  This is done by copying all records following the primary HDU from    */
/*  the input to the output.                                              */
/*   Inputs:                                                              */
/*      infile    input file name                                         */
/*      outfile   output file name                                        */
/*   Returns:                                                             */
/*      0 on success, else error                                          */
/*----------------------------------------------------------------------- */
{
  char buffer[2880], *ibuff;
  int  copy=0, bad=0;
  FILE *inFile, *outFile;
  size_t nread, nwrit;
  int i,first;

  if (!infile) return 1; /* pointer valid? */
  if (!outfile) return 1; /* pointer valid? */

  /* open files */
  first = 0; /* drop any initial '!' */
  if (infile[0]=='!') first=1;
  inFile = fopen (&infile[first], "r");
  if (inFile==NULL) { 
    fprintf(stderr,"append_tables: ERROR opening input file %s \n", 
	    infile);
    return 1;
  } /* end error check */

  /* output - open for append */
  first = 0; /* drop any initial '!' */
  if (infile[0]=='!') first=1;
  outFile = fopen (&outfile[first], "a");
  if (outFile==NULL) { 
    fprintf(stderr,"append_tables: ERROR opening output file %s \n", 
	    outfile);
    return 1;
  } /* end error check */

  /* loop over input file */
  while ((!ferror(inFile)) && (!feof(inFile))) {
    nread = fread (buffer, 2880, 1, inFile);
    if (feof(inFile)) break; 
    if (nread!=1) { /* problem */
      fprintf(stderr,"append_tables: ERROR reading input file %s \n", 
	      infile);
      return 1;
    }
    /* look for "END     " keyword */
    if (!copy) {
      ibuff = buffer;
      for (i=0; i<36; i++) { /* loop over header records */
	if (!strncmp("END     ", ibuff,8)) { /* found? */
	  copy = 1;
	}
	if (copy) break;
	ibuff +=80; /* update pointer */
      } /* end loop looking for "END     " */
    } else { /* copy record */
      nwrit = fwrite(buffer, 2880, 1, outFile);
      if (nwrit!=1) { /* problem */
	fprintf(stderr,"append_tables: ERROR writing output file %s \n", 
		outfile);
	return 1;
      }
    }
  } /* end loop over file */
  
  /* close files */
  fclose (inFile);
  fclose (outFile);
  return 0; /* finished OK */
} /* end append_tables */

void copy_table(char *table_type, fitsfile *infptr, fitsfile *outfptr, 
		int *ierr)
/*----------------------------------------------------------------------- */
/*  Copy a specified table from one open FITS file to another             */
/*   Inputs:                                                              */
/*      table_type table type name                                        */
/*      infptr     input file cfitsio structure                           */
/*      outfptr    output file  cfitsio structure, this file should be    */
/*                 explicitly closed to forse flushing the buffer.        */
/*   Outputs:                                                             */
/*      ierr       Error code, 0 = OK, else failed, checked on input      */
/*----------------------------------------------------------------------- */
{
  long irow, nrow;
  table_def *tabledef;
  table_row *row;

  if (*ierr>0) return;   /* existing error? */
  if (!infptr) return;   /* sanity check */
  if (!outfptr) return;  /* sanity check */

  /* position input FITS file at table */
  if (fits_movnam_hdu(infptr, BINARY_TBL, table_type, 0, ierr)) {
    fprintf(stderr,"copy_table: ERROR %d finding %s table \n", 
	    *ierr, table_type);
    return;
  }


  /* get input table definition */
  tabledef = make_table_def_from_table (infptr);
  if (!tabledef) { /* allocation or read error? */
    fprintf (stderr, "copy_table: ERROR reading input table def.\n");
    *ierr = 10000;
    return;
  }

  /* create 0utput table and open */
  make_table_from_def (outfptr, tabledef, ierr);
  if (*ierr!=0) { /* error? */
    fprintf (stderr, "copy_table: ERROR creating output table\n");
    kill_table_def(tabledef); /* cleanup */
    return;
  }

  /* create row structure */
  row = make_table_row_from_def (tabledef);
  if (!row) { /* allocation error? */
    fprintf (stderr, "copy_table: ERROR creating row structure.\n");
    *ierr = 10000;
    kill_table_def(tabledef); /* cleanup */
    return;
  }

  /* get size of table */
  fits_get_num_rows (infptr, &nrow, ierr);

  /* looping over rows copying */
  for (irow = 1; irow<=nrow; irow++) {
    row->trow = irow;
    table_read_row (infptr, tabledef, row, ierr); /* read */
    if (*ierr!=0) { /* error? */
      fprintf (stderr, "copy_table: ERROR %d reading input %s table\n",
	       *ierr, table_type);
      kill_table_def(tabledef); /* cleanup */
      kill_table_row(row); 
      return;
    }
    table_write_row (outfptr, tabledef, row, ierr); /* write */
    if (*ierr!=0) { /* error? */
      fprintf (stderr, "copy_table: ERROR %d writing output %s table\n",
	       *ierr, table_type);
      kill_table_def(tabledef); /* cleanup */
      kill_table_row(row); 
      return;
    }
  } /* end loop copying rows */

  /* cleanup */
  kill_table_def(tabledef);
  kill_table_row(row);
} /* end copy_table */

void copy_header_keywords(fitsfile *infptr, fitsfile *outfptr, 
		int nkeys, char** keys, int *ierr)
/*----------------------------------------------------------------------- */
/*  Copy all keywords in a specified list from one header to another.     */
/*  ESO Heirarchal keywords and specified list are copied.                */
/*   Inputs:                                                              */
/*      infptr     input file cfitsio structure                           */
/*      outfptr    output file  cfitsio structure                         */
/*      nkeys      number of entries in keys                              */
/*      keys       array of keywords to copy                              */
/*   Outputs:                                                             */
/*      ierr       Error code, 0 = OK, else failed, checked on input      */
/*----------------------------------------------------------------------- */
{
  int  i, ikey;
  int  wanted=0;
  char card[81], hierarch[]="HIERARCH ESO";
  int  keysexist, morekeys;
  
  if (*ierr>0) return;   /* existing error? */
  if (!infptr) return;   /* sanity check */
  if (!outfptr) return;  /* sanity check */
  
  /* How many are in the input file? */
  if (fits_get_hdrspace (infptr, &keysexist, &morekeys, ierr)) {
    fprintf(stderr,"copy_header_keywords: ERROR %d determining number of keywords in header \n", 
	    *ierr);
    return;
  }
  
  /* loop reading keywords */
  for (ikey=1; ikey<=keysexist; ikey++) {
    if (fits_read_record (infptr, ikey, card, ierr)) {
      fprintf(stderr,"copy_header_keywords: ERROR %d reading keyword %d \n", 
	      *ierr, ikey);
      return;
    }
    /* do we want this one? OK if either heirarch or in keys */
    wanted = !strncmp (hierarch, card, strlen(hierarch));
    if (!wanted) {
      for (i=0; i<nkeys; i++) {
	wanted = !strncmp (keys[i], card, strlen(keys[i]));
	if (wanted) break;
      } 
    }
      
    if (wanted) {
      if (fits_write_record (outfptr, card, ierr)) {
	fprintf(stderr,"copy_header_keywords: ERROR %d reading keyword %d \n", 
		*ierr, ikey);
	return;
      }
    }
  } /* end loop reading keywords from input */
} /* end copy_header_keywords */
  
int default_fits_read_key(fitsfile *infptr, int datatype, 
			  char *keyname, void* value, char *comment, 
			  void* deflt, int *status)
/*----------------------------------------------------------------------- */
/*  Attempt to read a given keyword but if it is not present return a     */
/*  default value.                                                        */
/*   Inputs:                                                              */
/*      infptr     input file cfitsio structure                           */
/*      datatype   cfitsio data type code                                 */
/*      keyname    keyword                                                */
/*      deflt      pointer to default value                               */
/*      nkeys      number of entries in keys                              */
/*      keys       array of keywords to copy                              */
/*   Outputs:                                                             */
/*      value      pointer to value                                       */
/*      comment    comment on keyword                                     */
/*      status     Error code, 0 = OK, else failed, checked on input      */
/*----------------------------------------------------------------------- */
{
  if (*status>0) return *status;   /* existing error? */
  if (!infptr) return -1;     /* sanity check */
   
  if (fits_read_key(infptr, datatype, keyname, value, comment, status)
      ==KEY_NO_EXIST) {  /* use default */
    *status = 0 ; /* clear error return */
    /* depends on datatype */
    switch (datatype) {
    case TSHORT:
      strncpy ((char*)value, (char*)deflt, sizeof(short));
      break;
    case TINT:
      strncpy ((char*)value, (char*)deflt, sizeof(int));
      break;
    case TLONG:
      strncpy ((char*)value, (char*)deflt, sizeof(long));
      break;
    case TSTRING:
      strncpy ((char*)value, (char*)deflt, strlen((char*)deflt));
      break;
    case TFLOAT:
      strncpy ((char*)value, (char*)deflt, sizeof(float));
      break;
    case TDOUBLE:
      strncpy ((char*)value, (char*)deflt, sizeof(double));
      break;
    default: /* unknown */
      strncpy ((char*)value, (char*)deflt, sizeof(short));
    } /* end switch finding datatype */

    /* blank comment */
    *comment = 0;
    /* clear error message stack */
    fits_clear_errmsg();

  } /* end of set default */
  return 0;
} /* end default_fits_read_key */

int default_fits_read_key_str(fitsfile *infptr, 
			      char *keyname, void* value, char *comment, 
			      void* deflt, int *status)
/*----------------------------------------------------------------------- */
/*  Attempt to read a string keyword but if it is not present return a    */
/*  default value.                                                        */
/*   Inputs:                                                              */
/*      infptr     input file cfitsio structure                           */
/*      keyname    keyword                                                */
/*      deflt      pointer to default value                               */
/*      nkeys      number of entries in keys                              */
/*      keys       array of keywords to copy                              */
/*   Outputs:                                                             */
/*      value      pointer to value                                       */
/*      comment    comment on keyword                                     */
/*      status     Error code, 0 = OK, else failed, checked on input      */
/*----------------------------------------------------------------------- */
{
  if (*status>0) return *status;   /* existing error? */
  if (!infptr) return -1;     /* sanity check */
   
  if (fits_read_key_str(infptr, keyname, value, comment, status)
      ==KEY_NO_EXIST) {  /* use default */
      strncpy ((char*)value, (char*)deflt, strlen((char*)deflt));
      *status = 0 ; /* clear error return */
    /* blank comment */
    *comment = 0;
    /* clear error message stack */
    fits_clear_errmsg();
  } /* end of set default */
  return 0;
} /* end default_fits_read_key_str */

int default_fits_read_key_dbl(fitsfile *infptr, 
			      char *keyname, void* value, char *comment, 
			      void* deflt, int *status)
/*----------------------------------------------------------------------- */
/*  Attempt to read a double keyword but if it is not present return a    */
/*  default value.                                                        */
/*   Inputs:                                                              */
/*      infptr     input file cfitsio structure                           */
/*      keyname    keyword                                                */
/*      deflt      pointer to default value                               */
/*      nkeys      number of entries in keys                              */
/*      keys       array of keywords to copy                              */
/*   Outputs:                                                             */
/*      value      pointer to value                                       */
/*      comment    comment on keyword                                     */
/*      status     Error code, 0 = OK, else failed, checked on input      */
/*----------------------------------------------------------------------- */
{
  if (*status>0) return *status;   /* existing error? */
  if (!infptr) return -1;     /* sanity check */
   
  if (fits_read_key_dbl(infptr, keyname, value, comment, status)
      ==KEY_NO_EXIST) {  /* use default */
    strncpy ((char*)value, (char*)deflt, sizeof(double));
    *status = 0 ; /* clear error return */
    /* blank comment */
    *comment = 0;
    /* clear error message stack */
    fits_clear_errmsg();
  } /* end of set default */
  return 0;
} /* end default_fits_read_key_dbl */

int default_fits_read_key_flt(fitsfile *infptr, 
			      char *keyname, void* value, char *comment, 
			      void* deflt, int *status)
/*----------------------------------------------------------------------- */
/*  Attempt to read a float keyword but if it is not present return a     */
/*  default value.                                                        */
/*   Inputs:                                                              */
/*      infptr     input file cfitsio structure                           */
/*      keyname    keyword                                                */
/*      deflt      pointer to default value                               */
/*      nkeys      number of entries in keys                              */
/*      keys       array of keywords to copy                              */
/*   Outputs:                                                             */
/*      value      pointer to value                                       */
/*      comment    comment on keyword                                     */
/*      status     Error code, 0 = OK, else failed, checked on input      */
/*----------------------------------------------------------------------- */
{
  if (*status>0) return *status;   /* existing error? */
  if (!infptr) return -1;     /* sanity check */
   
  if (fits_read_key_flt(infptr, keyname, value, comment, status)
      ==KEY_NO_EXIST) {  /* use default */
    strncpy ((char*)value, (char*)deflt, sizeof(float));
    *status = 0 ; /* clear error return */
    /* blank comment */
    *comment = 0;
    /* clear error message stack */
    fits_clear_errmsg();
  } /* end of set default */
  return 0;
} /* end default_fits_read_key_float */

int default_fits_read_key_lng(fitsfile *infptr, 
			      char *keyname, void* value, char *comment, 
			      void* deflt, int *status)
/*----------------------------------------------------------------------- */
/*  Attempt to read a long keyword but if it is not present return a      */
/*  default value.                                                        */
/*   Inputs:                                                              */
/*      infptr     input file cfitsio structure                           */
/*      keyname    keyword                                                */
/*      deflt      pointer to default value                               */
/*      nkeys      number of entries in keys                              */
/*      keys       array of keywords to copy                              */
/*   Outputs:                                                             */
/*      value      pointer to value                                       */
/*      comment    comment on keyword                                     */
/*      status     Error code, 0 = OK, else failed, checked on input      */
/*----------------------------------------------------------------------- */
{
  if (*status>0) return *status;   /* existing error? */
  if (!infptr) return -1;     /* sanity check */
   
  if (fits_read_key_lng(infptr, keyname, value, comment, status)
      ==KEY_NO_EXIST) {  /* use default */
    strncpy ((char*)value, (char*)deflt, sizeof(long));
    *status = 0 ; /* clear error return */
    /* blank comment */
    *comment = 0;
    /* clear error message stack */
    fits_clear_errmsg();
  } /* end of set default */
  return 0;
} /* end default_fits_read_key_lng */

float give_float_NaN()
/*----------------------------------------------------------------------- */
/*  return a float NaN (not a number)                                     */
/*  uses a word with all bits on                                          */
/*----------------------------------------------------------------------- */
{
  float TNaN, NaN;
  long *INaN = (long*)&TNaN; /*  Make a word with all bits on  */
  *INaN = ~0L;
  NaN = TNaN;
  return NaN;
}  /* end give_float_NaN */
/* */
